import stripe
from django.db import models

from .. import enums
from ..fields import JSONField, StripeEnumField, StripeForeignKey
from .base import StripeModel


class VerificationSession(StripeModel):
    """
    Stripe documentation: https://stripe.com/docs/api/identity/verification_sessions
    """

    stripe_class = stripe.identity.VerificationSession

    last_error = JSONField(null=True, blank=True)
    last_verification_report = StripeForeignKey(
        "djstripe.VerificationReport", on_delete=models.SET_NULL, null=True, blank=True
    )
    redaction = JSONField(null=True, blank=True)
    verified_outputs = JSONField(null=True, blank=True)
    status = StripeEnumField(enum=enums.VerificationSessionStatus)
    type = StripeEnumField(enum=enums.VerificationType)

    # The following attributes are not stored because they are sensitive.
    url = None
    client_secret = None


class VerificationReport(StripeModel):
    """
    Stripe documentation: https://stripe.com/docs/api/identity/verification_reports
    """

    stripe_class = stripe.identity.VerificationReport

    document = JSONField(null=True, blank=True)
    id_number = JSONField(null=True, blank=True)
    options = JSONField(null=True, blank=True)
    selfie = JSONField(null=True, blank=True)
    type = StripeEnumField(enum=enums.VerificationType)



"""
Django Administration interface definitions
"""
from typing import Any, Dict

from django.contrib import admin
from django.db import IntegrityError, transaction
from django.shortcuts import render
from stripe.error import InvalidRequestError

from djstripe import models

from .actions import CustomActionMixin
from .admin_inline import (
    InvoiceItemInline,
    LineItemInline,
    SubscriptionInline,
    SubscriptionItemInline,
    SubscriptionScheduleInline,
    TaxIdInline,
)
from .filters import CustomerHasSourceListFilter, CustomerSubscriptionStatusListFilter
from .forms import (
    APIKeyAdminCreateForm,
    CustomActionForm,
    WebhookEndpointAdminCreateForm,
    WebhookEndpointAdminEditForm,
)
from .utils import ReadOnlyMixin, get_forward_relation_fields_for_model


@admin.register(models.IdempotencyKey)
class IdempotencyKeyAdmin(ReadOnlyMixin, admin.ModelAdmin):
    list_display = ("uuid", "action", "created", "is_expired", "livemode")
    list_filter = ("livemode",)
    search_fields = ("uuid", "action")


@admin.register(models.WebhookEventTrigger)
class WebhookEventTriggerAdmin(ReadOnlyMixin, admin.ModelAdmin):
    list_display = (
        "created",
        "event",
        "stripe_trigger_account",
        "webhook_endpoint",
        "remote_ip",
        "processed",
        "valid",
        "exception",
        "djstripe_version",
    )
    list_filter = ("created", "valid", "processed")
    list_select_related = ("event",)
    raw_id_fields = get_forward_relation_fields_for_model(models.WebhookEventTrigger)

    def reprocess(self, request, queryset):
        for trigger in queryset:
            if not trigger.valid:
                self.message_user(request, "Skipped invalid trigger {trigger!r}")
                continue

            trigger.process()

    def get_queryset(self, request):
        return (
            super()
            .get_queryset(request)
            .select_related("stripe_trigger_account", "event", "webhook_endpoint")
        )


class StripeModelAdmin(CustomActionMixin, admin.ModelAdmin):
    """Base class for all StripeModel-based model admins"""

    change_form_template = "djstripe/admin/change_form.html"
    add_form_template = "djstripe/admin/add_form.html"
    actions = ("_resync_instances", "_sync_all_instances")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.raw_id_fields = get_forward_relation_fields_for_model(self.model)

    def get_list_display(self, request):
        return (
            ("__str__", "id", "djstripe_owner_account")
            + self.list_display
            + ("created", "livemode")
        )

    def get_list_filter(self, request):
        return self.list_filter + ("created", "livemode")

    def get_readonly_fields(self, request, obj=None):
        return self.readonly_fields + ("id", "djstripe_owner_account", "created")

    def get_search_fields(self, request):
        return self.search_fields + ("id",)

    def get_fieldsets(self, request, obj=None):
        common_fields = ("livemode", "id", "djstripe_owner_account", "created")
        # Have to remove the fields from the common set,
        # otherwise they'll show up twice.
        fields = [f for f in self.get_fields(request, obj) if f not in common_fields]
        return (
            (None, {"fields": common_fields}),
            (self.model.__name__, {"fields": fields}),
        )

    def get_queryset(self, request):
        return super().get_queryset(request).select_related("djstripe_owner_account")


@admin.register(models.Account)
class AccountAdmin(StripeModelAdmin):
    list_display = ("business_url", "country", "default_currency")
    list_filter = ("details_submitted",)
    search_fields = ("settings", "business_profile")


@admin.register(models.APIKey)
class APIKeyAdmin(admin.ModelAdmin):
    add_form_template = "djstripe/admin/add_form.html"
    change_form_template = "djstripe/admin/change_form.html"

    list_display = ("__str__", "type", "djstripe_owner_account", "livemode")
    readonly_fields = ("djstripe_owner_account", "livemode", "type", "secret")
    search_fields = ("name",)

    def get_readonly_fields(self, request, obj=None):
        if obj is None:
            return ["djstripe_owner_account", "livemode", "type"]
        return super().get_readonly_fields(request, obj=obj)

    def get_fields(self, request, obj=None):
        if obj is None:
            return APIKeyAdminCreateForm.Meta.fields
        return ["type", "djstripe_owner_account", "livemode", "name", "secret"]

    def get_form(self, request, obj=None, **kwargs):
        if obj is None:
            return APIKeyAdminCreateForm
        return super().get_form(request, obj, **kwargs)

    def get_queryset(self, request):
        return super().get_queryset(request).select_related("djstripe_owner_account")

    def save_model(self, request: Any, obj, form: Any, change: Any) -> None:
        try:
            # for non-existent Platform Accounts, because of Account._find_owner_account()
            # it will try to retrieve by api_key, Account.get_or_retrieve_for_api_key().
            # Account.get_or_retrieve_for_api_key will create this APIKey! This would cause
            # an IntegrityError as the APIKey gets created before this form gets saved
            with transaction.atomic():
                obj.save()
        except IntegrityError:
            pass


@admin.register(models.BalanceTransaction)
class BalanceTransactionAdmin(ReadOnlyMixin, StripeModelAdmin):
    list_display = (
        "type",
        "net",
        "amount",
        "fee",
        "currency",
        "available_on",
        "status",
    )
    list_filter = ("status", "type")


@admin.register(models.Charge)
class ChargeAdmin(StripeModelAdmin):
    list_display = (
        "customer",
        "amount",
        "invoice",
        "payment_method",
        "description",
        "paid",
        "disputed",
        "refunded",
        "fee",
    )

    search_fields = ("customer__id", "invoice__id")
    list_filter = ("status", "paid", "refunded", "captured")

    def get_queryset(self, request):
        return (
            super()
            .get_queryset(request)
            .select_related(
                "balance_transaction",
                "customer",
                "invoice",
                "payment_method",
                "payment_method__customer",
            )
        )


@admin.register(models.Coupon)
class CouponAdmin(StripeModelAdmin):
    list_display = (
        "amount_off",
        "percent_off",
        "duration",
        "duration_in_months",
        "redeem_by",
        "max_redemptions",
        "times_redeemed",
    )
    list_filter = ("duration", "redeem_by")
    radio_fields = {"duration": admin.HORIZONTAL}


@admin.register(models.Customer)
class CustomerAdmin(StripeModelAdmin):
    list_display = (
        "deleted",
        "subscriber",
        "email",
        "currency",
        "default_source",
        "default_payment_method",
        "coupon",
        "balance",
    )

    list_filter = (
        CustomerHasSourceListFilter,
        CustomerSubscriptionStatusListFilter,
        "deleted",
    )
    search_fields = ("email", "description", "deleted")
    inlines = (SubscriptionInline, SubscriptionScheduleInline, TaxIdInline)

    def get_queryset(self, request):
        return (
            super()
            .get_queryset(request)
            .select_related(
                "subscriber", "default_source", "default_payment_method", "coupon"
            )
        )


@admin.register(models.Discount)
class DiscountAdmin(ReadOnlyMixin, StripeModelAdmin):
    list_display = (
        "customer",
        "coupon",
        "invoice_item",
        "promotion_code",
        "subscription",
    )
    list_filter = ("customer", "start", "end", "promotion_code", "coupon")

    def get_actions(self, request):
        """
        Returns _resync_instances only for
        models with a defined model.stripe_class.retrieve
        """
        actions = super().get_actions(request)

        # remove "_sync_all_instances" as Discounts cannot be listed
        actions.pop("_sync_all_instances", None)

        return actions


@admin.register(models.Dispute)
class DisputeAdmin(ReadOnlyMixin, StripeModelAdmin):
    list_display = ("reason", "status", "amount", "currency", "is_charge_refundable")
    list_filter = ("is_charge_refundable", "reason", "status")


@admin.register(models.Event)
class EventAdmin(ReadOnlyMixin, StripeModelAdmin):
    list_display = ("type", "request_id")
    list_filter = ("type", "created")
    search_fields = ("request_id",)


@admin.register(models.File)
class FileAdmin(StripeModelAdmin):
    list_display = ("purpose", "size", "type")
    list_filter = ("purpose", "type")
    search_fields = ("filename",)


@admin.register(models.FileLink)
class FileLinkAdmin(StripeModelAdmin):
    list_display = ("url",)
    list_filter = ("expires_at",)

    def get_queryset(self, request):
        return super().get_queryset(request).select_related("file")


@admin.register(models.Order)
class OrderAdmin(StripeModelAdmin):
    list_display = (
        "amount_total",
        "customer",
        "status",
    )
    list_filter = (
        "customer",
        "status",
    )
    list_select_related = ("customer", "payment_intent")


@admin.register(models.PaymentIntent)
class PaymentIntentAdmin(StripeModelAdmin):
    list_display = (
        "on_behalf_of",
        "customer",
        "amount",
        "payment_method",
        "currency",
        "description",
        "amount_capturable",
        "amount_received",
        "receipt_email",
    )
    search_fields = ("customer__id", "invoice__id")

    def get_queryset(self, request):
        return (
            super()
            .get_queryset(request)
            .select_related(
                "customer", "payment_method", "payment_method__customer", "on_behalf_of"
            )
        )


@admin.register(models.Payout)
class PayoutAdmin(StripeModelAdmin):
    list_display = (
        "destination",
        "amount",
        "arrival_date",
        "method",
        "status",
        "type",
    )
    list_filter = ("destination__id", "original_payout")
    search_fields = ("destination__id", "balance_transaction__id")

    def get_queryset(self, request):
        return (
            super()
            .get_queryset(request)
            .select_related(
                "balance_transaction",
                "destination",
                "failure_balance_transaction",
                "original_payout",
                "reversed_by",
            )
        )


@admin.register(models.SetupIntent)
class SetupIntentAdmin(StripeModelAdmin):
    list_display = (
        "created",
        "customer",
        "description",
        "on_behalf_of",
        "payment_method",
        "payment_method_types",
        "status",
    )
    list_filter = ("status",)
    search_fields = ("customer__id", "status")

    def get_queryset(self, request):
        return (
            super()
            .get_queryset(request)
            .select_related(
                "on_behalf_of",
                "customer",
                "customer__subscriber",
                "payment_method",
                "payment_method__customer",
            )
        )


@admin.register(models.Session)
class SessionAdmin(StripeModelAdmin):
    list_display = ("customer", "customer_email", "subscription")
    list_filter = ("customer", "mode")
    search_fields = ("customer__id", "customer_email")

    def get_queryset(self, request):
        return super().get_queryset(request).select_related("customer", "subscription")


@admin.register(models.Invoice)
class InvoiceAdmin(StripeModelAdmin):
    list_display = (
        "total",
        "get_default_tax_rates",
        "paid",
        "currency",
        "number",
        "customer",
        "due_date",
    )
    list_filter = (
        "status",
        "paid",
        "attempted",
        "created",
        "due_date",
        "period_start",
        "period_end",
    )
    search_fields = ("customer__id", "number", "receipt_number")
    inlines = (InvoiceItemInline,)

    @admin.display(description="Default Tax Rates")
    def get_default_tax_rates(self, obj):
        result = [str(tax_rate) for tax_rate in obj.default_tax_rates.all()]
        if result:
            return ", ".join(result)

    def get_queryset(self, request):
        return (
            super()
            .get_queryset(request)
            .select_related("customer")
            .prefetch_related("default_tax_rates")
        )


@admin.register(models.LineItem)
class LineItemAdmin(StripeModelAdmin):
    list_display = (
        "amount",
        "invoice_item",
        "subscription",
        "subscription_item",
        "type",
    )
    list_filter = ("type", "invoice_item", "subscription", "subscription_item")
    list_select_related = ("invoice_item", "subscription", "subscription_item")


@admin.register(models.Mandate)
class MandateAdmin(StripeModelAdmin):
    list_display = ("status", "type", "payment_method")
    list_filter = ("multi_use", "single_use")
    search_fields = ("payment_method__id",)

    def get_queryset(self, request):
        return super().get_queryset(request).select_related("payment_method")

    def get_actions(self, request):
        """
        Returns _resync_instances only for
        models with a defined model.stripe_class.retrieve
        """
        actions = super().get_actions(request)

        # remove "_sync_all_instances" as Mandates cannot be listed
        actions.pop("_sync_all_instances", None)

        return actions


@admin.register(models.Plan)
class PlanAdmin(StripeModelAdmin):
    radio_fields = {"interval": admin.HORIZONTAL}

    def get_readonly_fields(self, request, obj=None):
        """Return extra readonly_fields."""
        readonly_fields = super().get_readonly_fields(request, obj)

        if obj:
            readonly_fields += (
                "amount",
                "currency",
                "interval",
                "interval_count",
                "trial_period_days",
            )

        return readonly_fields

    def get_queryset(self, request):
        return (
            super()
            .get_queryset(request)
            .select_related("product")
            .prefetch_related("subscriptions")
        )


@admin.register(models.Price)
class PriceAdmin(StripeModelAdmin):
    list_display = ("product", "currency", "active")
    list_filter = ("active", "type", "billing_scheme", "tiers_mode")
    raw_id_fields = ("product",)
    search_fields = ("nickname",)
    radio_fields = {"type": admin.HORIZONTAL}

    def get_queryset(self, request):
        return (
            super()
            .get_queryset(request)
            .select_related("product")
            .prefetch_related("product__prices")
        )


@admin.register(models.Product)
class ProductAdmin(StripeModelAdmin):
    list_display = (
        "name",
        "default_price",
        "type",
        "active",
        "url",
        "statement_descriptor",
    )
    list_filter = ("type", "active", "shippable")
    search_fields = ("name", "statement_descriptor")

    def get_queryset(self, request):
        return super().get_queryset(request).prefetch_related("prices")


@admin.register(models.Refund)
class RefundAdmin(StripeModelAdmin):
    list_display = (
        "amount",
        "currency",
        "charge",
        "reason",
        "status",
        "failure_reason",
    )
    list_filter = ("reason", "status")
    search_fields = ("receipt_number",)

    def get_queryset(self, request):
        return super().get_queryset(request).select_related("charge")


@admin.register(models.Source)
class SourceAdmin(StripeModelAdmin):
    list_display = ("customer", "type", "status", "amount", "currency", "usage", "flow")
    list_filter = ("type", "status", "usage", "flow")

    def get_queryset(self, request):
        return (
            super()
            .get_queryset(request)
            .select_related("customer", "customer__subscriber")
        )


@admin.register(models.SourceTransaction)
class SourceTransactionAdmin(StripeModelAdmin):
    list_display = ("status", "amount", "currency")
    list_filter = (
        "status",
        "source__id",
        "source__customer",
        "source__customer__subscriber",
    )
    list_select_related = ("source", "source__customer", "source__customer__subscriber")


@admin.register(models.PaymentMethod)
class PaymentMethodAdmin(StripeModelAdmin):
    list_display = ("customer", "type", "billing_details")
    list_filter = ("type",)
    search_fields = ("customer__id",)

    def get_queryset(self, request):
        return (
            super()
            .get_queryset(request)
            .select_related("customer", "customer__subscriber")
        )


@admin.register(models.Card)
class CardAdmin(StripeModelAdmin):
    list_display = ("customer", "account")
    search_fields = ("customer__id", "account__id")

    def get_queryset(self, request):
        return (
            super()
            .get_queryset(request)
            .select_related(
                "customer",
                "customer__default_source",
                "customer__default_payment_method",
                "account",
            )
        )


@admin.register(models.BankAccount)
class BankAccountAdmin(StripeModelAdmin):
    list_display = ("customer", "account")
    search_fields = ("customer__id", "account__id")

    def get_queryset(self, request):
        return (
            super()
            .get_queryset(request)
            .select_related(
                "customer",
                "customer__default_source",
                "customer__default_payment_method",
                "account",
            )
        )


@admin.register(models.ShippingRate)
class ShippingRateAdmin(StripeModelAdmin):
    list_display = ("display_name", "active", "tax_behavior", "tax_code")
    list_filter = ("active", "tax_behavior")
    list_select_related = ("tax_code",)


@admin.register(models.Subscription)
class SubscriptionAdmin(StripeModelAdmin):
    list_display = ("customer", "status", "get_product_name", "get_default_tax_rates")
    list_filter = ("status", "cancel_at_period_end")

    inlines = (SubscriptionItemInline, SubscriptionScheduleInline, LineItemInline)

    def get_actions(self, request):
        # get all actions
        actions = super().get_actions(request)
        actions["_cancel"] = self.get_action("_cancel")
        return actions

    @admin.action(description="Cancel selected subscriptions")
    def _cancel(self, request, queryset):
        """Cancel a subscription."""
        context = self.get_admin_action_context(queryset, "_cancel", CustomActionForm)
        return render(request, "djstripe/admin/confirm_action.html", context)

    def get_queryset(self, request):
        return (
            super()
            .get_queryset(request)
            .select_related(
                "customer",
                "plan",
                "plan__product",
            )
            .prefetch_related(
                "customer__subscriptions",
                "customer__subscriptions__plan",
                "customer__subscriptions__plan__product",
                "default_tax_rates",
            )
        )

    @admin.display(description="Default Tax Rates")
    def get_default_tax_rates(self, obj):
        result = [str(tax_rate) for tax_rate in obj.default_tax_rates.all()]
        if result:
            return ", ".join(result)

    @admin.display(description="Product Name")
    def get_product_name(self, obj):
        if obj.plan and obj.plan.product:
            return obj.plan.product.name


@admin.register(models.SubscriptionSchedule)
class SubscriptionScheduleAdmin(StripeModelAdmin):
    list_display = ("status", "subscription", "current_phase", "customer")
    list_filter = ("status", "subscription", "customer")
    list_select_related = ("customer", "customer__subscriber", "subscription")

    @admin.display(description="Release Selected Subscription Schedules")
    def _release_subscription_schedule(self, request, queryset):
        """Release a SubscriptionSchedule."""
        context = self.get_admin_action_context(
            queryset, "_release_subscription_schedule", CustomActionForm
        )
        return render(request, "djstripe/admin/confirm_action.html", context)

    @admin.display(description="Cancel Selected Subscription Schedules")
    def _cancel_subscription_schedule(self, request, queryset):
        """Cancel a SubscriptionSchedule."""
        context = self.get_admin_action_context(
            queryset, "_cancel_subscription_schedule", CustomActionForm
        )
        return render(request, "djstripe/admin/confirm_action.html", context)

    def get_actions(self, request):
        # get all actions
        actions = super().get_actions(request)
        actions["_release_subscription_schedule"] = self.get_action(
            "_release_subscription_schedule"
        )
        actions["_cancel_subscription_schedule"] = self.get_action(
            "_cancel_subscription_schedule"
        )
        return actions


@admin.register(models.TaxCode)
class TaxCodeAdmin(StripeModelAdmin):
    list_display = ("name", "description")
    list_filter = ("name",)


@admin.register(models.TaxRate)
class TaxRateAdmin(StripeModelAdmin):
    list_display = ("active", "display_name", "inclusive", "jurisdiction", "percentage")
    list_filter = ("active", "inclusive", "jurisdiction")


@admin.register(models.Transfer)
class TransferAdmin(StripeModelAdmin):
    list_display = ("amount", "description")


@admin.register(models.TransferReversal)
class TransferReversalAdmin(StripeModelAdmin):
    list_display = ("amount", "transfer")


@admin.register(models.ApplicationFee)
class ApplicationFeeAdmin(StripeModelAdmin):
    list_display = ("amount", "account")


@admin.register(models.ApplicationFeeRefund)
class ApplicationFeeReversalAdmin(StripeModelAdmin):
    list_display = ("amount", "fee")


@admin.register(models.UsageRecord)
class UsageRecordAdmin(StripeModelAdmin):
    list_display = ("quantity", "subscription_item", "timestamp")

    def get_queryset(self, request):
        return super().get_queryset(request).select_related("subscription_item")

    def get_actions(self, request):
        """
        Returns _resync_instances only for
        models with a defined model.stripe_class.retrieve
        """
        actions = super().get_actions(request)

        # remove "_sync_all_instances" as UsageRecords cannot be listed
        actions.pop("_sync_all_instances", None)

        return actions


@admin.register(models.UsageRecordSummary)
class UsageRecordSummaryAdmin(StripeModelAdmin):
    list_display = ("invoice", "subscription_item", "total_usage")

    def get_queryset(self, request):
        return (
            super().get_queryset(request).select_related("invoice", "subscription_item")
        )


@admin.register(models.WebhookEndpoint)
class WebhookEndpointAdmin(CustomActionMixin, admin.ModelAdmin):
    change_form_template = "djstripe/admin/webhook_endpoint/change_form.html"
    delete_confirmation_template = (
        "djstripe/admin/webhook_endpoint/delete_confirmation.html"
    )
    add_form_template = "djstripe/admin/webhook_endpoint/add_form.html"
    readonly_fields = ("url",)
    list_display = (
        "__str__",
        "djstripe_owner_account",
        "livemode",
        "created",
        "api_version",
    )
    actions = ("_resync_instances", "_sync_all_instances")

    def get_actions(self, request):
        actions = super().get_actions(request)
        # Disable the mass-delete action for webhook endpoints.
        # We don't want to enable deleting multiple endpoints on Stripe at once.
        if "delete_selected" in actions:
            del actions["delete_selected"]
        return actions

    def get_form(self, request, obj=None, **kwargs):
        if obj:
            return WebhookEndpointAdminEditForm
        return WebhookEndpointAdminCreateForm

    def get_readonly_fields(self, request, obj=None):
        if obj:
            return (
                "id",
                "livemode",
                "api_version",
                "url",
                "created",
                "djstripe_owner_account",
                "djstripe_uuid",
            )
        return super().get_readonly_fields(request, obj=obj)

    def get_fieldsets(self, request, obj=None):
        if obj:
            # if djstripe_uuid is null, this is not a dj-stripe webhook
            header_fields = ["id", "livemode", "djstripe_owner_account", "url"]
            advanced_fields = [
                "enabled_events",
                "metadata",
                "api_version",
                "djstripe_uuid",
                "tolerance",
            ]
            if obj.djstripe_uuid:
                core_fields = [
                    "enabled",
                    "base_url",
                    "description",
                ]
            else:
                core_fields = [
                    "enabled",
                    "description",
                ]
        else:
            header_fields = ["djstripe_owner_account", "livemode"]
            core_fields = ["description", "base_url", "connect"]
            advanced_fields = [
                "metadata",
                "api_version",
                "enabled_events",
                "tolerance",
            ]

        return [
            (None, {"fields": header_fields}),
            ("Endpoint configuration", {"fields": core_fields}),
            (
                "Advanced",
                {"fields": advanced_fields, "classes": ["collapse"]},
            ),
        ]

    def get_changeform_initial_data(self, request) -> Dict[str, str]:
        ret = super().get_changeform_initial_data(request)
        base_url = f"{request.scheme}://{request.get_host()}"
        ret.setdefault("base_url", base_url)
        return ret

    def delete_model(self, request, obj: models.WebhookEndpoint):
        try:
            obj._api_delete()
        except InvalidRequestError as e:
            if e.user_message.startswith("No such webhook endpoint: "):
                # Webhook was already deleted in Stripe
                pass
            else:
                raise

        return super().delete_model(request, obj)

    def get_queryset(self, request):
        return super().get_queryset(request).select_related("djstripe_owner_account")



"""
Django Administration Utils Module
"""


class ReadOnlyMixin:
    def has_add_permission(self, request):
        return False

    def has_change_permission(self, request, obj=None):
        return False


def get_forward_relation_fields_for_model(model):
    """Return an iterable of the field names that are forward relations,
    I.E ManyToManyField, OneToOneField, and ForeignKey.

    Useful for perhaps ensuring the admin is always using raw ID fields for
    newly added forward relation fields.
    """
    return [
        field.name
        for field in model._meta.get_fields()
        # Get only relation fields
        if field.is_relation
        # Exclude auto relation fields, like reverse one to one.
        and not field.auto_created
        # We only want forward relations.
        and any((field.many_to_many, field.one_to_one, field.many_to_one))
    ]



"""
Django Administration Custom Actions Module
"""
from django.contrib import admin
from django.contrib.admin import helpers
from django.contrib.admin.utils import quote
from django.shortcuts import render
from django.urls import path, reverse
from django.utils.html import format_html
from django.utils.text import capfirst

from . import views
from .forms import CustomActionForm


class CustomActionMixin:
    # So that actions get shown even if there are 0 instances
    # https://docs.djangoproject.com/en/dev/ref/contrib/admin/#django.contrib.admin.ModelAdmin.show_full_result_count
    show_full_result_count = False

    def get_urls(self):
        custom_urls = [
            path(
                "action/<str:action_name>/<str:model_name>/",
                self.admin_site.admin_view(views.ConfirmCustomAction.as_view()),
                name="djstripe_custom_action",
            ),
        ]
        return custom_urls + super().get_urls()

    def get_admin_action_context(self, queryset, action_name, form_class):
        context = {
            "action_name": action_name,
            "model_name": self.model._meta.model_name,
            "info": [],
            "queryset": queryset,
            "changelist_url": reverse(
                f"admin:{self.model._meta.app_label}_{self.model._meta.model_name}_changelist"
            ),
            "ACTION_CHECKBOX_NAME": helpers.ACTION_CHECKBOX_NAME,
            "form": form_class(
                initial={
                    helpers.ACTION_CHECKBOX_NAME: queryset.values_list("pk", flat=True)
                },
                model_name=self.model._meta.model_name,
                action_name=action_name,
            ),
        }

        if action_name == "_sync_all_instances":
            context["form"] = form_class(
                initial={helpers.ACTION_CHECKBOX_NAME: [action_name]},
                model_name=self.model._meta.model_name,
                action_name=action_name,
            )

        else:
            for obj in queryset:
                admin_url = reverse(
                    f"admin:{obj._meta.app_label}_{obj._meta.model_name}_change",
                    None,
                    (quote(obj.pk),),
                )
                context["info"].append(
                    format_html(
                        '{}: <a href="{}">{}</a>',
                        capfirst(obj._meta.verbose_name),
                        admin_url,
                        obj,
                    )
                )
        return context

    def get_actions(self, request):
        """
        Returns _resync_instances only for
        models with a defined model.stripe_class.retrieve
        """
        actions = super().get_actions(request)

        # ensure we return "_resync_instances" ONLY for
        # models that have a GET method
        if not getattr(self.model.stripe_class, "retrieve", None):
            actions.pop("_resync_instances", None)

        return actions

    @admin.action(description="Re-Sync Selected Instances")
    def _resync_instances(self, request, queryset):
        """Admin Action to resync selected instances"""
        context = self.get_admin_action_context(
            queryset, "_resync_instances", CustomActionForm
        )
        return render(request, "djstripe/admin/confirm_action.html", context)

    @admin.action(description="Sync All Instances for all API Keys")
    def _sync_all_instances(self, request, queryset):
        """Admin Action to Sync All Instances"""
        context = self.get_admin_action_context(
            queryset, "_sync_all_instances", CustomActionForm
        )
        return render(request, "djstripe/admin/confirm_action.html", context)

    def changelist_view(self, request, extra_context=None):
        # we fool it into thinking we have selected some query
        # since we need to sync all instances
        post = request.POST.copy()
        if (
            helpers.ACTION_CHECKBOX_NAME not in post
            and post.get("action") == "_sync_all_instances"
        ):
            post[helpers.ACTION_CHECKBOX_NAME] = None
            request._set_post(post)
        return super().changelist_view(request, extra_context)



"""
Utility functions related to the djstripe app.
"""
import datetime
from typing import Optional

import stripe
from django.apps import apps
from django.conf import settings
from django.contrib.humanize.templatetags.humanize import intcomma
from django.db.models.query import QuerySet
from django.utils import timezone


def get_supported_currency_choices(api_key):
    """
    Pull a stripe account's supported currencies and returns a choices tuple of those
    supported currencies.

    :param api_key: The api key associated with the account from which to pull data.
    :type api_key: str
    """
    account = stripe.Account.retrieve(api_key=api_key)
    supported_payment_currencies = stripe.CountrySpec.retrieve(
        account["country"], api_key=api_key
    )["supported_payment_currencies"]

    return [(currency, currency.upper()) for currency in supported_payment_currencies]


def clear_expired_idempotency_keys():
    from .models import IdempotencyKey

    threshold = timezone.now() - datetime.timedelta(hours=24)
    IdempotencyKey.objects.filter(created__lt=threshold).delete()


def convert_tstamp(response) -> Optional[datetime.datetime]:
    """
    Convert a Stripe API timestamp response (unix epoch) to a native datetime.
    """
    if response is None:
        # Allow passing None to convert_tstamp()
        return response

    # Overrides the set timezone to UTC - I think...
    tz = get_timezone_utc() if settings.USE_TZ else None

    return datetime.datetime.fromtimestamp(response, tz)


# TODO: Finish this.
CURRENCY_SIGILS = {"CAD": "$", "EUR": "€", "GBP": "£", "USD": "$"}


def get_friendly_currency_amount(amount, currency: str) -> str:
    currency = currency.upper()
    sigil = CURRENCY_SIGILS.get(currency, "")
    amount_two_decimals = f"{amount:.2f}"
    return f"{sigil}{intcomma(amount_two_decimals)} {currency}"


class QuerySetMock(QuerySet):
    """
    A mocked QuerySet class that does not handle updates.
    Used by UpcomingInvoice.invoiceitems (deprecated) and UpcomingInvoice.lineitems.
    """

    @classmethod
    def from_iterable(cls, model, iterable):
        instance = cls(model)
        instance._result_cache = list(iterable)
        instance._prefetch_done = True
        return instance

    def _clone(self):
        return self.__class__.from_iterable(self.model, self._result_cache)

    def update(self):
        return 0

    def delete(self):
        return 0


def get_id_from_stripe_data(data):
    """
    Extract stripe id from stripe field data
    """

    if isinstance(data, str):
        # data like "sub_6lsC8pt7IcFpjA"
        return data
    elif data:
        # data like {"id": sub_6lsC8pt7IcFpjA", ...}
        return data.get("id")
    else:
        return None


def get_model(model_name):
    app_label = "djstripe"
    app_config = apps.get_app_config(app_label)
    model = app_config.get_model(model_name)
    return model


def get_queryset(pks, model_name):
    model = get_model(model_name)
    return model.objects.filter(pk__in=pks)


def get_timezone_utc():
    """
    Returns UTC attribute in a backwards compatible way.

    UTC attribute has been moved from django.utils.timezone module to
    datetime.timezone class
    """
    try:
        # Django 4+
        return datetime.timezone.utc
    except AttributeError:
        return timezone.utc



"""
signals are sent for each event Stripe sends to the app

Stripe docs for Webhooks: https://stripe.com/docs/webhooks
"""
from django.dispatch import Signal

# providing_args=["instance", "api_key"]
webhook_pre_validate = Signal()

# providing_args=["instance", "api_key", "valid"]
webhook_post_validate = Signal()

# providing_args=["instance", "api_key"]
webhook_pre_process = Signal()

# providing_args=["instance", "api_key"]
webhook_post_process = Signal()

# providing_args=["instance", "api_key", "exception", "data"]
webhook_processing_error = Signal()

ENABLED_EVENTS = [
    # Update this by copy-pasting the "enabled_events" enum values from
    # https://raw.githubusercontent.com/stripe/openapi/master/openapi/spec3.json
    "*",
    "account.application.authorized",
    "account.application.deauthorized",
    "account.external_account.created",
    "account.external_account.deleted",
    "account.external_account.updated",
    "account.updated",
    "application_fee.created",
    "application_fee.refund.updated",
    "application_fee.refunded",
    "balance.available",
    "billing_portal.configuration.created",
    "billing_portal.configuration.updated",
    "billing_portal.session.created",
    "capability.updated",
    "cash_balance.funds_available",
    "charge.captured",
    "charge.dispute.closed",
    "charge.dispute.created",
    "charge.dispute.funds_reinstated",
    "charge.dispute.funds_withdrawn",
    "charge.dispute.updated",
    "charge.expired",
    "charge.failed",
    "charge.pending",
    "charge.refund.updated",
    "charge.refunded",
    "charge.succeeded",
    "charge.updated",
    "checkout.session.async_payment_failed",
    "checkout.session.async_payment_succeeded",
    "checkout.session.completed",
    "checkout.session.expired",
    "coupon.created",
    "coupon.deleted",
    "coupon.updated",
    "credit_note.created",
    "credit_note.updated",
    "credit_note.voided",
    "customer.created",
    "customer.deleted",
    "customer.discount.created",
    "customer.discount.deleted",
    "customer.discount.updated",
    "customer.source.created",
    "customer.source.deleted",
    "customer.source.expiring",
    "customer.source.updated",
    "customer.subscription.created",
    "customer.subscription.deleted",
    "customer.subscription.paused",
    "customer.subscription.pending_update_applied",
    "customer.subscription.pending_update_expired",
    "customer.subscription.resumed",
    "customer.subscription.trial_will_end",
    "customer.subscription.updated",
    "customer.tax_id.created",
    "customer.tax_id.deleted",
    "customer.tax_id.updated",
    "customer.updated",
    "customer_cash_balance_transaction.created",
    "file.created",
    "financial_connections.account.created",
    "financial_connections.account.deactivated",
    "financial_connections.account.disconnected",
    "financial_connections.account.reactivated",
    "financial_connections.account.refreshed_balance",
    "identity.verification_session.canceled",
    "identity.verification_session.created",
    "identity.verification_session.processing",
    "identity.verification_session.redacted",
    "identity.verification_session.requires_input",
    "identity.verification_session.verified",
    "invoice.created",
    "invoice.deleted",
    "invoice.finalization_failed",
    "invoice.finalized",
    "invoice.marked_uncollectible",
    "invoice.paid",
    "invoice.payment_action_required",
    "invoice.payment_failed",
    "invoice.payment_succeeded",
    "invoice.sent",
    "invoice.upcoming",
    "invoice.updated",
    "invoice.voided",
    "invoiceitem.created",
    "invoiceitem.deleted",
    "invoiceitem.updated",
    "issuing_authorization.created",
    "issuing_authorization.request",
    "issuing_authorization.updated",
    "issuing_card.created",
    "issuing_card.updated",
    "issuing_cardholder.created",
    "issuing_cardholder.updated",
    "issuing_dispute.closed",
    "issuing_dispute.created",
    "issuing_dispute.funds_reinstated",
    "issuing_dispute.submitted",
    "issuing_dispute.updated",
    "issuing_transaction.created",
    "issuing_transaction.updated",
    "mandate.updated",
    "order.created",
    "payment_intent.amount_capturable_updated",
    "payment_intent.canceled",
    "payment_intent.created",
    "payment_intent.partially_funded",
    "payment_intent.payment_failed",
    "payment_intent.processing",
    "payment_intent.requires_action",
    "payment_intent.succeeded",
    "payment_link.created",
    "payment_link.updated",
    "payment_method.attached",
    "payment_method.automatically_updated",
    "payment_method.detached",
    "payment_method.updated",
    "payout.canceled",
    "payout.created",
    "payout.failed",
    "payout.paid",
    "payout.reconciliation_completed",
    "payout.updated",
    "person.created",
    "person.deleted",
    "person.updated",
    "plan.created",
    "plan.deleted",
    "plan.updated",
    "price.created",
    "price.deleted",
    "price.updated",
    "product.created",
    "product.deleted",
    "product.updated",
    "promotion_code.created",
    "promotion_code.updated",
    "quote.accepted",
    "quote.canceled",
    "quote.created",
    "quote.finalized",
    "radar.early_fraud_warning.created",
    "radar.early_fraud_warning.updated",
    "recipient.created",
    "recipient.deleted",
    "recipient.updated",
    "refund.created",
    "refund.updated",
    "reporting.report_run.failed",
    "reporting.report_run.succeeded",
    "reporting.report_type.updated",
    "review.closed",
    "review.opened",
    "setup_intent.canceled",
    "setup_intent.created",
    "setup_intent.requires_action",
    "setup_intent.setup_failed",
    "setup_intent.succeeded",
    "sigma.scheduled_query_run.created",
    "sku.created",
    "sku.deleted",
    "sku.updated",
    "source.canceled",
    "source.chargeable",
    "source.failed",
    "source.mandate_notification",
    "source.refund_attributes_required",
    "source.transaction.created",
    "source.transaction.updated",
    "subscription_schedule.aborted",
    "subscription_schedule.canceled",
    "subscription_schedule.completed",
    "subscription_schedule.created",
    "subscription_schedule.expiring",
    "subscription_schedule.released",
    "subscription_schedule.updated",
    "tax_rate.created",
    "tax_rate.updated",
    "terminal.reader.action_failed",
    "terminal.reader.action_succeeded",
    "test_helpers.test_clock.advancing",
    "test_helpers.test_clock.created",
    "test_helpers.test_clock.deleted",
    "test_helpers.test_clock.internal_failure",
    "test_helpers.test_clock.ready",
    "topup.canceled",
    "topup.created",
    "topup.failed",
    "topup.reversed",
    "topup.succeeded",
    "transfer.created",
    "transfer.reversed",
    "transfer.updated",
    "treasury.credit_reversal.created",
    "treasury.credit_reversal.posted",
    "treasury.debit_reversal.completed",
    "treasury.debit_reversal.created",
    "treasury.debit_reversal.initial_credit_granted",
    "treasury.financial_account.closed",
    "treasury.financial_account.created",
    "treasury.financial_account.features_status_updated",
    "treasury.inbound_transfer.canceled",
    "treasury.inbound_transfer.created",
    "treasury.inbound_transfer.failed",
    "treasury.inbound_transfer.succeeded",
    "treasury.outbound_payment.canceled",
    "treasury.outbound_payment.created",
    "treasury.outbound_payment.expected_arrival_date_updated",
    "treasury.outbound_payment.failed",
    "treasury.outbound_payment.posted",
    "treasury.outbound_payment.returned",
    "treasury.outbound_transfer.canceled",
    "treasury.outbound_transfer.created",
    "treasury.outbound_transfer.expected_arrival_date_updated",
    "treasury.outbound_transfer.failed",
    "treasury.outbound_transfer.posted",
    "treasury.outbound_transfer.returned",
    "treasury.received_credit.created",
    "treasury.received_credit.failed",
    "treasury.received_credit.succeeded",
    "treasury.received_debit.created",
    # special case? - TODO can be deleted?
    "ping",
]


# A signal for each Event type. See https://stripe.com/docs/api/events/types

WEBHOOK_SIGNALS = dict(
    [  # providing_args=["event"]
        (hook, Signal()) for hook in ENABLED_EVENTS if hook != "*"
    ]
)



import stripe
from django.db import models

from djstripe.settings import djstripe_settings

from .. import enums
from ..fields import (
    JSONField,
    StripeCurrencyCodeField,
    StripeEnumField,
    StripeForeignKey,
    StripeQuantumCurrencyAmountField,
)
from .base import StripeModel


class Session(StripeModel):
    """
    A Checkout Session represents your customer's session as they pay
    for one-time purchases or subscriptions through Checkout.

    Stripe documentation: https://stripe.com/docs/api/checkout/sessions?lang=python
    """

    expand_fields = [
        "line_items",
        "total_details.breakdown",
        "line_items.data.discounts",
    ]
    stripe_class = stripe.checkout.Session

    amount_total = StripeQuantumCurrencyAmountField(
        null=True,
        blank=True,
        help_text="Total of all items after discounts and taxes are applied.",
    )
    amount_subtotal = StripeQuantumCurrencyAmountField(
        null=True,
        blank=True,
        help_text="Total of all items after discounts and taxes are applied.",
    )
    billing_address_collection = StripeEnumField(
        enum=enums.SessionBillingAddressCollection,
        blank=True,
        help_text=(
            "The value (auto or required) for whether Checkout"
            "collected the customer's billing address."
        ),
    )
    cancel_url = models.TextField(
        max_length=5000,
        blank=True,
        help_text=(
            "The URL the customer will be directed to if they"
            "decide to cancel payment and return to your website."
        ),
    )
    client_reference_id = models.TextField(
        max_length=5000,
        blank=True,
        help_text=(
            "A unique string to reference the Checkout Session."
            "This can be a customer ID, a cart ID, or similar, and"
            "can be used to reconcile the session with your internal systems."
        ),
    )
    currency = StripeCurrencyCodeField(
        null=True,
        blank=True,
        help_text=(
            "Three-letter ISO currency code, in lowercase. Must be a supported"
            " currency."
        ),
    )
    customer = StripeForeignKey(
        "Customer",
        null=True,
        on_delete=models.SET_NULL,
        help_text="Customer this Checkout is for if one exists.",
    )
    customer_email = models.CharField(
        max_length=255,
        blank=True,
        help_text=(
            "If provided, this value will be used when the Customer object is created."
        ),
    )
    display_items = JSONField(
        null=True,
        blank=True,
        help_text="The line items, plans, or SKUs purchased by the customer.",
    )
    line_items = JSONField(
        null=True,
        blank=True,
        help_text="The line items purchased by the customer.",
    )
    locale = models.CharField(
        max_length=255,
        blank=True,
        help_text=(
            "The IETF language tag of the locale Checkout is displayed in."
            "If blank or auto, the browser's locale is used."
        ),
    )
    mode = StripeEnumField(
        enum=enums.SessionMode,
        blank=True,
        help_text=(
            "The mode of the Checkout Session, one of payment, setup, or subscription."
        ),
    )
    payment_intent = StripeForeignKey(
        "PaymentIntent",
        null=True,
        on_delete=models.SET_NULL,
        help_text="PaymentIntent created if SKUs or line items were provided.",
    )
    payment_method_types = JSONField(
        help_text=(
            "The list of payment method types (e.g. card) that this "
            "Checkout Session is allowed to accept."
        )
    )
    payment_status = StripeEnumField(
        enum=enums.SessionPaymentStatus,
        null=True,
        blank=True,
        help_text=(
            "The payment status of the Checkout Session, one of paid, unpaid, or"
            " no_payment_required. You can use this value to decide when to fulfill"
            " your customer's order."
        ),
    )
    setup_intent = StripeForeignKey(
        "SetupIntent",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        help_text="The ID of the SetupIntent for Checkout Sessions in setup mode.",
    )
    shipping_address_collection = JSONField(
        null=True,
        blank=True,
        help_text=(
            "When set, provides configuration for Checkout to collect a shipping"
            " address from a customer."
        ),
    )
    shipping_cost = JSONField(
        null=True,
        blank=True,
        help_text=(
            "The details of the customer cost of shipping, including the customer"
            " chosen ShippingRate."
        ),
    )
    shipping_details = JSONField(
        null=True,
        blank=True,
        help_text="Shipping information for this Checkout Session.",
    )
    shipping_options = JSONField(
        null=True,
        blank=True,
        help_text="The shipping rate options applied to this Session.",
    )
    status = StripeEnumField(
        enum=enums.SessionStatus,
        null=True,
        blank=True,
        help_text=(
            "The status of the Checkout Session, one of open, complete, or expired."
        ),
    )
    submit_type = StripeEnumField(
        enum=enums.SubmitTypeStatus,
        blank=True,
        help_text=(
            "Describes the type of transaction being performed by Checkoutin order to"
            " customize relevant text on the page, such as the submit button."
        ),
    )
    subscription = StripeForeignKey(
        "Subscription",
        null=True,
        on_delete=models.SET_NULL,
        help_text="Subscription created if one or more plans were provided.",
    )
    success_url = models.TextField(
        max_length=5000,
        blank=True,
        help_text=(
            "The URL the customer will be directed to after the payment or subscription"
            "creation is successful."
        ),
    )
    total_details = JSONField(
        null=True,
        blank=True,
        help_text="Tax and discount details for the computed total amount.",
    )
    url = models.TextField(
        max_length=5000,
        blank=True,
        null=True,
        help_text=(
            "The URL to the Checkout Session. Redirect customers to this URL to take"
            " them to Checkout. If you’re using Custom Domains, the URL will use your"
            " subdomain. Otherwise, it’ll use checkout.stripe.com. This value is only"
            " present when the session is active."
        ),
    )

    def _attach_objects_post_save_hook(
        self,
        cls,
        data,
        api_key=djstripe_settings.STRIPE_SECRET_KEY,
        pending_relations=None,
    ):
        from ..event_handlers import update_customer_helper

        super()._attach_objects_post_save_hook(
            cls, data, api_key=api_key, pending_relations=pending_relations
        )

        # only update if customer and metadata exist
        if self.customer and self.metadata:
            key = djstripe_settings.SUBSCRIBER_CUSTOMER_KEY
            current_value = self.metadata.get(key)

            # only update if metadata has the SUBSCRIBER_CUSTOMER_KEY
            if current_value:
                metadata = {key: current_value}

                # Update the customer with ONLY the customer specific metadata
                update_customer_helper(
                    metadata,
                    self.customer.id,
                    key,
                )

                # Update metadata in the Upstream Customer Object on Stripe
                self.customer._api_update(metadata=metadata)



"""
Django Administration Inline interface definitions
"""
from django.contrib import admin

from djstripe import models

from .utils import get_forward_relation_fields_for_model


class SubscriptionInline(admin.StackedInline):
    """A TabularInline for use models.Subscription."""

    model = models.Subscription
    extra = 0
    readonly_fields = ("id", "created", "djstripe_owner_account")
    raw_id_fields = get_forward_relation_fields_for_model(model)
    show_change_link = True


class SubscriptionScheduleInline(admin.StackedInline):
    """A TabularInline for use models.SubscriptionSchedule."""

    model = models.SubscriptionSchedule
    extra = 0
    readonly_fields = ("id", "created", "djstripe_owner_account")
    raw_id_fields = get_forward_relation_fields_for_model(model)
    show_change_link = True

    def __init__(self, parent_model, admin_site):
        super().__init__(parent_model, admin_site)

        # dynamically set fk_name as SubscriptionScheduleInline is used
        # in CustomerAdmin as well as SubscriptionAdmin
        if parent_model is models.Subscription:
            self.fk_name = "subscription"


class TaxIdInline(admin.TabularInline):
    """A TabularInline for use models.Subscription."""

    model = models.TaxId
    extra = 0
    max_num = 5
    readonly_fields = (
        "id",
        "created",
        "verification",
        "livemode",
        "country",
        "djstripe_owner_account",
    )
    show_change_link = True


class SubscriptionItemInline(admin.StackedInline):
    """A TabularInline for use models.Subscription."""

    model = models.SubscriptionItem
    extra = 0
    readonly_fields = ("id", "created", "djstripe_owner_account")
    raw_id_fields = get_forward_relation_fields_for_model(model)
    show_change_link = True


class InvoiceItemInline(admin.StackedInline):
    """A TabularInline for use InvoiceItem."""

    model = models.InvoiceItem
    extra = 0
    readonly_fields = ("id", "created", "djstripe_owner_account")
    raw_id_fields = get_forward_relation_fields_for_model(model)
    show_change_link = True


class LineItemInline(admin.StackedInline):
    """A TabularInline for LineItem."""

    model = models.LineItem
    extra = 0
    readonly_fields = ("id", "created", "djstripe_owner_account")
    raw_id_fields = get_forward_relation_fields_for_model(model)
    show_change_link = True
